/**
 * JASMINe
 * Copyright (C) 2009-2011 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: RuleEngineService.java 7874 2011-03-28 09:50:05Z danesa $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.rules.common;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.osgi.framework.ServiceReference;
import org.ow2.jasmine.monitoring.eos.common.osgi.OSGiManager;
import org.ow2.jasmine.rules.manager.model.IRulePackage;
import org.ow2.jasmine.rules.manager.model.textual.ITextualRulePackage;
import org.ow2.jasmine.rules.manager.model.visual.ILHSElement;
import org.ow2.jasmine.rules.manager.model.visual.IRHSElement;
import org.ow2.jasmine.rules.manager.model.visual.IRule;
import org.ow2.jasmine.rules.manager.model.visual.IVisualRulePackage;
import org.ow2.jasmine.rules.manager.model.visual.LogicalExpression;
import org.ow2.jasmine.rules.manager.model.visual.Rule;
import org.ow2.jasmine.rules.manager.model.visual.VisualRulePackage;
import org.ow2.jasmine.rules.manager.service.DroolsManagerFactory;
import org.ow2.jasmine.rules.manager.service.IDroolsManager;
import org.ow2.jasmine.rules.manager.service.IDroolsManagerFactory;
import org.ow2.jasmine.rules.manager.service.RulePackageCompilationError;
import org.ow2.jasmine.rules.manager.service.RulePackageSaveReport;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Rule engine server-side service. Only one running instance of this service is
 * needed because it manage several rule engine instances à the same time.
 * Provides all the methods needed by the client side rules-related components.
 *
 * @author Julien Nicoulaud
 * @see {@link org.ow2.jasmine.rules.manager.service.IDroolsManager}
 * @see {@link org.ow2.jasmine.rules.manager.service.IDroolsManagerFactory}
 */
public class RuleEngineService implements Serializable {

    // =========================================================================
    // Variables
    // =========================================================================

    /**
     * Serial version ID
     */
    private static final long serialVersionUID = 1943812711958376631L;

    /**
     * Logger
     */
    private static Log logger = LogFactory.getLog(RuleEngineService.class);

    /**
     * Temporary directory where files are stored
     */
    private final String tempDir = System.getProperty("java.io.tmpdir") + File.separatorChar + "EoS.temp" + File.separatorChar;

    /**
     * Files read/write buffer size
     */
    private static final int DEFAULT_BUFFER_SIZE = 2048;

    /**
     * The OSGi manager
     */
    OSGiManager osgiManager = new OSGiManager();

    /**
     * The manager factory
     */
    IDroolsManagerFactory managerFactory;

    /**
     * The Drools Managers
     */
    private Map<String,IDroolsManager> droolsManagers =
        new HashMap<String,IDroolsManager>();


    // =========================================================================
    // Public methods - client side provided API
    // =========================================================================

    // Rule engine control

    /**
     * Start the rule engine
     */
    public void start(final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Start the rule engine
        ruleEngineManager.start();
    }

    /**
     * Stop the rule engine
     */
    public void stop(final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Start the rule engine
        ruleEngineManager.stop();
    }


    // Status & Statistics

    /**
     * Get the rule engine status
     */
    public RuleEngineStatus getRuleEngineStatus(final String ruleEngineName)
        throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Build a Rule Engine Status object
        RuleEngineStatus res = new RuleEngineStatus();

        //// Set the name
        res.ruleEngineName = ruleEngineName;

        //// Set the state
        res.engineStatus = ruleEngineManager.getState().toString();

        //// Set some details
        switch(ruleEngineManager.getState()) {
            case IDLE:
                break;
            case RUNNING:
                List<String> deployedPackagesNames = ruleEngineManager.getDeployedPackagesNames();
                if (deployedPackagesNames != null) {
                    res.nbDeployedPackages = deployedPackagesNames.size();
                    Iterator<String> it = deployedPackagesNames.iterator();
                    for (int i = 0 ; it.hasNext(); i++) {
                        if (i == 0) {
                            res.deployedPackages = it.next();
                        } else {
                            String nextPackage = ", " + it.next();
                            res.deployedPackages.concat(nextPackage);
                        }
                    }
                }
                break;
            case STOPPED:
                break;
        }

        return res;
    }

    /**
     * Get all the rule packages associated to a given rule engine
     *
     * @param ruleEngineName
     * @return
     * @throws Exception
     */
    public List<IRulePackage> getRulePackages(final String ruleEngineName)
        throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Get the packages
        IRulePackage[] res = ruleEngineManager.getRulePackages();

        // Return the packages
        if (res != null) {
            return Arrays.asList(res);
        }
        return new ArrayList<IRulePackage>();
    }

    /**
     * Get a rule package identified by its unique systemName attribute
     *
     * @param rulePackageSystemName
     * @param ruleEngineName
     * @return
     */
    public IRulePackage getRulePackage(final String rulePackageSystemName,
                                       final String ruleEngineName) throws Exception {


        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Return the package
        return ruleEngineManager.getRulePackage(rulePackageSystemName);
    }

    /**
     * Commit a rule package changes
     *
     * @param rulePackage The rule package to save
     * @param ruleEngineName The target rule engine instance
     * @return A "save" operation report to inform precisely the client side
     * @throws Exception
     */
    public RulePackageSaveReport saveRulePackage(final IRulePackage rulePackage,
                                                    final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Apply changes
        return ruleEngineManager.commitRulePackageChanges(rebuildPackage(rulePackage));
    }

    /**
     * Delete a rule package from the database
     *
     * @param rulePackage The rule package to save
     * @param ruleEngineName The target rule engine instance
     * @throws Exception
     */
    public void deleteRulePackage(final IRulePackage rulePackage,
                                  final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Apply changes
        ruleEngineManager.deleteRulePackage(rebuildPackage(rulePackage));
    }

    /**
     * Test if a rule package compiles well
     * @param rulePackage
     * @param ruleEngineName
     * @throws Exception
     */
    public List<RulePackageCompilationError> compileRulePackage(final IRulePackage rulePackage,
                                                                final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Run the service command and return the result
        return ruleEngineManager.checkCompilation(rebuildPackage(rulePackage));
    }

    /**
     * Test if a rule package name is available
     * @param systemName the name to check availability
     * @param ruleEngineName
     * @throws Exception
     */
    public Boolean checkNameAvailability(final String systemName,
                                         final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Run the service command and return the result
        return ruleEngineManager.checkNameAvailability(systemName);
    }


    // Packages binding XML <=> Java

    /**
     * Serialize a collection of rule packages to XML
     */
    public Document exportRulePackagesAsXML(final List<IRulePackage> rulePackages,
                                          final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Initialize the temporary directory
        File tmpDirFile = new File(tempDir);
        tmpDirFile.mkdirs();

        // Format the date inserted in file names
        SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");
        String date = dateFormat.format(new Date());

        // Prepare the XML response to the client
        DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document fileToDownloadInfo = docBuilder.newDocument();
        Element fileToDownloadInfoNode = fileToDownloadInfo.createElement("file");
        fileToDownloadInfo.appendChild(fileToDownloadInfoNode);

        // Create a file for each serialized rule package
        List<File> files = new ArrayList<File>();

        for (IRulePackage rulePackage : rulePackages) {

            // Write it to a file
            File file = new File(tempDir + rulePackage.getSystemName().substring(rulePackage.getSystemName().lastIndexOf(".")+1) + "_" + date + ".xml");
            FileOutputStream fileOutStream = new FileOutputStream(file);
            fileOutStream.write(ruleEngineManager.exportRulePackageToXML(rebuildPackage(rulePackage)).getBytes());
            fileOutStream.close();

            // Store it in our list
            files.add(file);
        }

        // If several rule packages uploaded, create an archive
        if (files.size() > 1) {

            // Create a ZIP file
            File zipFile = new File(tempDir + "XML_packages_" + date + ".zip");
            BufferedInputStream origin = null;
            FileOutputStream dest = new FileOutputStream(zipFile);
            ZipOutputStream zipOut = new ZipOutputStream(new BufferedOutputStream(dest));
            byte data[] = new byte[DEFAULT_BUFFER_SIZE];

            // Add each rule package file, at the ZIP file root folder
            for (int i=0; i<files.size(); i++) {
               FileInputStream fi = new FileInputStream(files.get(i));
               origin = new BufferedInputStream(fi, DEFAULT_BUFFER_SIZE);
               ZipEntry entry = new ZipEntry(files.get(i).getName());
               zipOut.putNextEntry(entry);
               int count;
               while((count = origin.read(data, 0, DEFAULT_BUFFER_SIZE)) != -1) {
                  zipOut.write(data, 0, count);
               }
               origin.close();
            }
            zipOut.close();

            // Build the response
            fileToDownloadInfoNode.setTextContent(zipFile.getAbsolutePath());
            fileToDownloadInfoNode.setAttribute("name",zipFile.getName());
            fileToDownloadInfoNode.setAttribute("size",String.valueOf(zipFile.length()/1024));
        }

        // If only one rule package uploaded, send directly the link to
        // the corresponding file
        else {
            fileToDownloadInfoNode.setTextContent(files.get(0).getAbsolutePath());
            fileToDownloadInfoNode.setAttribute("name",files.get(0).getName());
            fileToDownloadInfoNode.setAttribute("size",String.valueOf(files.get(0).length()/1024));
        }
        return fileToDownloadInfo;
    }

    /**
     * Build a rule package from its XML serialized representation
     *
     * @param filePath the path to the file to import
     * @return
     * @throws Exception
     */
    public IRulePackage importRulePackageFromXML(final String filePath,
                                                 final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Read the file
        StringBuilder fileContent = new StringBuilder();
        BufferedReader input =  new BufferedReader(new FileReader(new File(filePath)));
        String line = null;
        while (( line = input.readLine()) != null){
            fileContent.append(line);
            fileContent.append(System.getProperty("line.separator"));
        }
        input.close();

        return ruleEngineManager.importRulePackageFromXML(fileContent.toString().trim());
    }


    // =========================================================================
    // Private methods
    // =========================================================================

    /**
     * Get or create the rule engine manager associated to a rule engine instance
     */
    private IDroolsManager getRuleEngineManager(final String ruleEngineName)
        throws Exception {

        // If no registered manager, retrieve it
        if(!droolsManagers.containsKey(ruleEngineName)){

            try {

                // Get the manager factory service reference
                ServiceReference managerFactoryRef =
                    osgiManager.getBundleContext().getServiceReference(IDroolsManagerFactory.class.getName());

                // Get its service instance
                managerFactory = (DroolsManagerFactory) osgiManager.getBundleContext().getService(managerFactoryRef);

                // Get a manager from it and store it
                droolsManagers.put(ruleEngineName,managerFactory.getManager(ruleEngineName));

            } catch (Exception e) {
                throw new Exception("Could not connect to the rule engine manager",e);
            }
        }

        // Return the manager
        return droolsManagers.get(ruleEngineName);
    }

    /**
     * Rebuild manually a given rule package. This is a workaround for the fact
     * that Flex builds the Java objects with its own subtypes (for instance,
     * flex.messaging.io.ArrayCollection that extends ArrayList).
     *
     * FIXME : Instead of doing this, Granite's jar that contains flex.* classes
     * should be externalized from JASMINe EoS and accessible in the bundles that
     * need it...
     *
     * @param rulePackage the package to rebuild
     * @return the package built with no flex.* class
     */
    private IRulePackage rebuildPackage(final IRulePackage rulePackage) {
        if (rulePackage instanceof ITextualRulePackage) {
            return rulePackage;
        }
        if (rulePackage instanceof IVisualRulePackage) {
            IVisualRulePackage convertedRP = new VisualRulePackage();
            convertedRP.setDescription(rulePackage.getDescription());
            convertedRP.setDisplayName(rulePackage.getDisplayName());
            convertedRP.setPreviousSystemName(rulePackage.getPreviousSystemName());
            convertedRP.setRuleEngineInstanceName(rulePackage.getRuleEngineInstanceName());
            convertedRP.setState(rulePackage.getState());
            convertedRP.setSystemName(rulePackage.getSystemName());

            ArrayList<String> convertedImports = new ArrayList<String>();
            for (String importStatement : ((IVisualRulePackage) rulePackage).getImports()) {
                convertedImports.add(importStatement);
            }
            convertedRP.setImports(convertedImports);

            ArrayList<IRule> convertedRules = new ArrayList<IRule>();
            for (IRule rule : ((IVisualRulePackage)rulePackage).getRules()) {
                IRule convertedRule = new Rule();
                convertedRule.setName(rule.getName());

                ArrayList<String> convertedVariables = new ArrayList<String>();
                for (String variable : rule.getVariables()) {
                    convertedImports.add(variable);
                }
                convertedRule.setVariables(convertedVariables);

                ArrayList<ILHSElement> convertedLHS = new ArrayList<ILHSElement>();
                for (ILHSElement lhsElement : rule.getLHS()) {
                    convertedLHS.add(rebuildLHSElement(lhsElement));
                }
                convertedRule.setLHS(convertedLHS);

                ArrayList<IRHSElement> convertedRHS = new ArrayList<IRHSElement>();
                for (IRHSElement rhsElement : rule.getRHS()) {
                    convertedRHS.add(rebuildRHSElement(rhsElement));
                }
                convertedRule.setRHS(convertedRHS);

                convertedRules.add(convertedRule);
            }
            convertedRP.setRules(convertedRules);

            return convertedRP;
        }
        return null;
    }

    /**
     * Rebuild a ILHSElement
     * @see {@link RuleEngineService#rebuildPackage(IRulePackage)}
     */
    private ILHSElement rebuildLHSElement(final ILHSElement element) {

        if (element instanceof LogicalExpression) {
            LogicalExpression convertedLogicalExpression = new LogicalExpression();
            convertedLogicalExpression.setOperator(((LogicalExpression) element).getOperator());

            ArrayList<ILHSElement> convertedOperands = new ArrayList<ILHSElement>();
            for (ILHSElement lhsElement : ((LogicalExpression) element).getOperands()) {
                convertedOperands.add(rebuildLHSElement(lhsElement));
            }
            convertedLogicalExpression.setOperands(convertedOperands);
            return convertedLogicalExpression;
        }
        return element;
    }

    /**
     * Rebuild a IRHSElement
     * @see {@link RuleEngineService#rebuildPackage(IRulePackage)}
     */
    private IRHSElement rebuildRHSElement(final IRHSElement element) {
        return element;
    }

}
